@@ -3,10 +3,30 @@
 package Mail::Summary::Tools::Output::HTML;
 use Moose;
 
-use Mail::Summary::Tools::Output::TT;
+use HTML::Element;
 use Text::Markdown ();
 use HTML::Entities;
 
+use utf8;
+
+has body_only => (
+	isa => "Bool",
+	is  => "rw",
+	default => 0,
+);
+
+has strip_divs => (
+	isa => "Bool",
+	is  => "rw",
+	default => 0,
+);
+
+has lang => (
+	isa => "Str",
+	is  => "rw",
+	default => "en",
+);
+
 has summary => (
 	isa => "Mail::Summary::Tools::Summary",
 	is  => "rw",
@@ -30,28 +50,65 @@ has generator => (
 	},
 );
 
-has template_config => (
-	isa => "HashRef",
-	is  => "rw",
-	default => sub { return {} },
-);
-
-has template_obj => (
-	isa => "Template",
-	is  => "rw",
-	lazy => 1,
-	default => sub { Template->new( $_[0]->template_config ) },
-);
-
 sub process {
 	my $self = shift;
 
-	$self->template_snippet(
-		$self->main_template,
-		html => $self,
+	my @tree = HTML::Element->new_from_lol( $self->body_only ? $self->body : $self->document_structure );
+
+	@tree = $self->scrub(@tree);
+
+	$self->emit(@tree);
+
+}
+
+sub document_structure {
+	my $self = shift;
+
+	return (
+		[html => { xmlns => "http://www.w3.org/1999/xhtml", 'xml:lang' => $self->lang },
+			[head =>
+				[title => $self->summary->title ],
+				[meta => { 'http-equiv' => "Content-Type", content => "text/html; charset=utf-8" }],
+				[meta => { name => "description", content => $self->description }],
+				[meta => { name => "generator",   content => $self->generator }],
+				$self->css,
+			],
+			[body => 
+				$self->toc,
+				$self->body,
+			],
+		
+		]
 	);
 }
 
+sub scrub {
+	my ( $self, @tree ) = @_;
+
+	if ( $self->strip_divs ) {
+		@tree = $self->scrub_strip_divs(@tree);
+	}
+
+	return @tree;
+}
+
+sub scrub_strip_divs {
+	my ( $self, @tree ) = @_;
+
+	foreach my $subtree ( @tree ) {
+		foreach my $div ( $subtree->find_by_tag_name('div') ) {
+			$div->replace_with_content if defined $div->parent;
+		}
+	}
+
+	map { ($_->tag eq "div") ? $_->content_list : $_ } @tree;
+}
+
+sub emit {
+	my ( $self, @tree ) = @_;
+	return @tree;
+}
+
 sub template_snippet {
 	my ( $self, $snippet, %vars ) = @_;
 	
@@ -77,11 +134,11 @@ sub markdown {
 	$text =~ s/<((?:msgid|rt):\S+?)>/$self->expand_uri($1)/ge;
 	$text =~ s/\[(.*?)\]\(((?:msgid|rt):\S+?)\)/$self->expand_uri($2, $1)/ge;
 
-	# non ascii stuff gets escaped (accents, etc), but not punctuation, which
-	# markdown will handle for us.
-	$text = $self->escape_html($text, '^\p{IsASCII}');
+	my $html = Text::Markdown::markdown( $text );
 
-	Text::Markdown::markdown( $text );
+	# non ascii stuff gets escaped (accents, etc), but not punctuation, which
+	# markdown will handle for us
+	['~literal' => { text => $self->escape_unicode($html) } ];
 }
 
 sub rt_uri {
@@ -131,10 +188,25 @@ sub expand_uri {
 	}
 }
 
+sub escape_unicode {
+	my ( $self, $text ) = @_;
+	$self->escape_html($text, '^\p{IsASCII}');
+}
+
 sub escape_html {
 	my ( $self, $text, @extra ) = @_;
 	HTML::Entities::encode_entities($text, @extra);
-}	
+}
+
+sub div {
+	my ( $self, $class_spec, @elems ) = @_;
+
+	my $class_attr = (ref $class_spec
+		? join(" ", @$class_spec)
+		: $class_spec );
+
+	[ div => { class => $class_attr }, @elems ];
+}
 
 has h1_tag => (
 	isa => "ArrayRef",
@@ -144,13 +216,13 @@ has h1_tag => (
 );
 
 sub wrap_tags {
-	my ( $self, $tags, @text ) = @_;
+	my ( $self, $tags, @elems ) = @_;
 
 	if ( @$tags ) {
 		my ( $outer, @inner ) = @$tags;
-		return "<$outer>" . $self->wrap_tags( \@inner, @text ) . "</$outer>";
+		return [ $outer => $self->wrap_tags( \@inner, @elems ) ];
 	} else {
-		return "@text";
+		return @elems;
 	}
 }
 
@@ -194,93 +266,84 @@ sub toc {
 sub body {
 	my $self = shift;
 
-	return qq{<div id="summary_container">\n}
-	. join("\n\n\n",
+	return [ div => { id => "summary_container" },
 		$self->header,
 		$self->lists,
 		$self->footer,
-	)
-	. qq{</div>\n};
+	];
 }
 
 sub header {
 	my $self = shift;
 	my @parts;
 	
-	return qq{<div id="summary_header">\n}
-	. join("\n\n",
-		$self->h1( $self->escape_html( $self->summary->title || "Mailing list summary" ) ),
+	return [ div => { id => "summary_header" },
+		$self->h1( $self->summary->title || "Mailing list summary" ),
 		$self->custom_header,
-	)
-	. qq{</div>\n};
+	];
 }
 
 sub custom_header {
 	my $self = shift;
 
 	if ( my $header = eval { $self->summary->extra->{header} } ) {
-		return join("\n\n", map { $self->custom_header_section( $_ ) } @$header );
+		return ( map { $self->custom_header_section( $_ ) } @$header );
 	} else {
-		return ();
+		return;
 	}
 }
 
 sub custom_header_section {
 	my ( $self, $section ) = @_;
-	return qq{<div class="header_section">\n}
-	. $self->generic_custom_section( $section )
-	. qq{</div>\n};
+
+	return $self->div( header_section => $self->generic_custom_section( $section ) );
 }
 
 sub footer {
 	my $self = shift;
 	
-	return qq{<div id="summary_footer">\n}
-	. join("\n\n",
+	return [ div => { id => "summary_footer" },
 		$self->custom_footer,
 		$self->see_also,
-	)
-	. qq{</div>\n};
+	];
 }
 
 sub custom_footer {
 	my $self = shift;
 
 	if ( my $footer = eval { $self->summary->extra->{footer} } ) {
-		return join("\n\n", map { $self->custom_footer_section( $_ ) } @$footer );
+		return ( map { $self->custom_footer_section( $_ ) } @$footer );
 	} else {
-		return ();
+		return;
 	}
 }
 
 sub custom_footer_section {
 	my ( $self, $section ) = @_;
-	return qq{<div class="footer_section">\n}
-	. $self->generic_custom_section( $section )
-	. qq{</div>\n}
+	return $self->div( footer_section => 
+		$self->generic_custom_section( $section ),
+	);
 }
 
 sub generic_custom_section {
 	my ( $self, $section ) = @_;
 
-	return join("\n\n",
-		$self->h2( $self->escape_html($section->{title}) ),
+	return (
+		$self->h2( $section->{title} ),
 		$self->markdown( $section->{body} ),
-	)
+	);
 }
 
 sub see_also {
 	my $self = shift;
 
 	if ( my $see_also = eval { $self->summary->extra->{see_also} } ) {
-		return qq{<div class="footer_section" id="see_also">\n}
-		. join("\n\n",
+		return [ div => { id => "see_also", class => "footer_section" },
 			$self->see_also_heading($see_also),
 			$self->see_also_links($see_also),
-		)
-		. qq{\n</div>\n};
+		];
 	} else {
-		return ();
+		return;
 	}
 }
 
@@ -291,46 +354,44 @@ sub see_also_heading {
 
 sub see_also_links {
 	my ( $self, $see_also ) = @_;	
-
-	join("\n",
-		"<ul>",
-		( map { "<li>".$self->see_also_link($_)."</li>" } @$see_also ),
-		"</ul>",
-	);
+	[ ul => map { [ li => $self->see_also_link($_) ] } @$see_also ];
 }
 
 sub see_also_link {
 	my ( $self, $item ) = @_;
-	sprintf '<a href="%s">%s</a>', map { $self->escape_html($_) } $item->{uri}, $item->{name};
+	[a => { href => $item->{uri} }, $item->{name} ];
 }
 
 sub lists {
 	my $self = shift;
 
-	return qq{<div id="summary_container_body">\n}
-	. join("\n\n\n", map { $self->list($_) } $self->summary->lists)
-	. qq{</div>\n};
+	return $self->div( summary_container_body => 
+		map { $self->list($_) } $self->summary->lists
+	);
 }
 
 sub list {
 	my ( $self, $list ) = @_;
 
-	my $name = $list->name;
-	$name =~ s/[^\w]+/_/g;
+	( my $id = $list->name ) =~ s/[^\w]+/_/g;
+	
+	my @body = $self->list_body($list);
 
-	return qq{<div class="summary_list" id="summary_list_$name">\n}
-	. join("\n\n",
-		$self->list_header($list),
-		$self->list_body($list),
-		$self->list_footer($list),
-	)
-	. qq{</div>\n};
+	if ( @body ) {
+		return [ div => { id => "summay_list_$id", class => 'summary_list' },
+			$self->list_header($list),
+			@body,
+			$self->list_footer($list),
+		];
+	} else {
+		return;
+	}
 }
 
 sub list_header {
 	my ( $self, $list ) = @_;
 
-	return join("\n\n",
+	return (
 		$self->list_heading($list),
 		$self->list_description($list),
 	);
@@ -338,7 +399,10 @@ sub list_header {
 
 sub list_heading {
 	my ( $self, $list ) = @_;
-	$self->h2( $self->list_title($list) || return, $self->list_heading_extra );
+
+	my $title = $self->list_title($list) || return;
+
+	$self->h2( $title, $self->list_heading_extra($list) );
 }
 
 sub list_heading_extra {
@@ -353,9 +417,9 @@ sub list_title {
 	my $title = $list->title || $list->name || return;
 
 	if ( my $uri = eval { $list->extra->{uri} } ) {
-		return sprintf '<a href="%s">%s</a>', map { $self->escape_html($_) } $uri, $title,
+		return [a => { href => $uri }, $title ],
 	} else {
-		return $self->escape_html($title);
+		return $title,
 	}
 }
 
@@ -372,12 +436,15 @@ sub list_description {
 sub list_body {
 	my ( $self, $list ) = @_;
 
-	my $name = $list->name;
-	$name =~ s/[^\w]+/_/g;
+	( my $id = $list->name ) =~ s/[^\w]+/_/g;
 
-	return qq{<div class="summary_list_body" id="summary_list_body_$name">\n}
-	. join("\n\n", map { $self->thread($_) } $list->threads)
-	. qq{</div>\n};
+	if ( my @threads = map { $self->thread($_) } $list->threads ) {
+		return [ div => { id => "summary_list_body_$id", class => 'summary_list_body'  },
+			@threads,
+		];
+	} else {
+		return;
+	}
 }
 
 sub list_footer {
@@ -390,13 +457,11 @@ sub thread {
 
 	return if $thread->hidden;
 
-	return qq{<div class="thread_summary">\n}
-	. join("\n\n",
+	return $self->div( thread_summary =>
 		$self->thread_header($thread),
 		$self->thread_body($thread),
 		$self->thread_footer($thread),
-	)
-	. qq{</div>\n};
+	);
 }
 
 sub thread_header {
@@ -409,20 +474,19 @@ sub thread_link {
 
 	my $uri = $thread->archive_link->thread_uri;
 
-	sprintf '<a href="%s" name="%s">%s</a>', map { $self->escape_html($_) } $uri, $thread->message_id, $thread->subject,
+	[a => { href => $uri, name=> $thread->message_id }, $thread->subject ],
 }
 
 sub thread_body {
 	my ( $self, $thread ) = @_;
 
 	if ( my $summary = $thread->summary ) {
-		return qq{<div class="thread_summary_body">\n}
-		. $self->markdown($summary)
-		. qq{</div>\n};
+		return $self->div( thread_summary_body => $self->markdown($summary) );
 	} else {
-		return qq{<div class="thread_summary_body empty_thread_summary_body">\n}
-		. $self->thread_body_no_summary($thread)
-		. qq{</div>\n};
+		return $self->div(
+			[qw/thread_summary_body empty_thread_summary_body/],
+			$self->thread_body_no_summary($thread),
+		);
 	}
 }
 
@@ -431,8 +495,8 @@ sub thread_body_no_summary {
 
 	my $posters = eval { $thread->extra->{posters} };
 
-	join("\n",
-		'<p>No summary provided.</p>',
+	return (
+		[p => 'No summary provided.' ],
 		($posters ? $self->thread_posters($posters) : ()),
 	);
 }
@@ -440,11 +504,9 @@ sub thread_body_no_summary {
 sub thread_posters {
 	my ( $self, $posters ) = @_;
 
-	join("\n",
-		"<p>The following people participated in this thread:</p>",
-		"<ul>",
-		( map { "<li>" . $self->escape_html($_->{name} || $_->{email}) . "</li>" } @$posters ),
-		"</ul>",
+	return (
+		[p => "The following people participated in this thread:" ],
+		[ul => map { [li => ['~literal' => { text => $self->escape_unicode($_->{name} || $_->{email}) } ] ] } @$posters ],
 	);
 }
 
@@ -458,23 +520,6 @@ sub css {
 	return ();
 }
 
-use constant main_template => <<'TMPL';
-<?xml version="1.0" encoding="UTF-8"?>
-<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
-<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
-	<head>
-		<title>[% summary.title | html %]</title>
-		<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
-		<meta name="description" content="[% html.description | html %]" />
-		<meta name="generator" content="[% html.generator | html %]" />
-		[% html.css %]
-	</head>
-	<body>
-		[% html.toc %]
-		[% html.body %]
-	</body>
-</html>
-TMPL
 
 __PACKAGE__;
 
